/******************************************************************************
 * Copyright 2022 Daniel Dietrich
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { Fn } from "./functions";
import { Is, Or, UnionToIntersection } from "./utilities";
export declare type IsEmpty<T> = [keyof T] extends [never] ? true : false;
export declare type Obj = Record<PropertyKey, unknown>;
export declare type Paths<T, P = TupledPaths<T>> = P extends [string, unknown] ? {
    [K in `${P[0]}`]: P[1];
} : never;
declare type TupledPaths<T, K = keyof T> = T extends Obj ? K extends string | number ? T[K] extends Record<string | number, unknown> ? TupledPaths<T[K]> extends infer F ? F extends [string, unknown] ? [`${K}.${F[0]}`, F[1]] : never : never : [`${K}`, T[K]] : never : never;
export declare type Map<T = any, R = unknown> = [unknown] extends [T] ? unknown : R;
export declare type FlatMap<T = any, R = unknown> = Flatten<Map<T, R>>;
export declare type Flatten<T extends Map> = Join<UnionToIntersection<T>>;
export declare type Join<T> = T extends Obj ? {
    [K in keyof T]: T[K];
} : T;
export declare type Filter<T, V> = T extends any[] ? FilterArray<T, V> : T extends Obj ? FilterObj<T, V> : never;
export declare type FilterNot<T, V> = T extends any[] ? FilterNotArray<T, V> : T extends Obj ? FilterNotObj<T, V> : never;
declare type FilterObj<T, V> = Pick<T, {
    [K in keyof T]-?: T[K] extends V ? K : never;
}[keyof T]>;
declare type FilterNotObj<T, V> = Pick<T, {
    [K in keyof T]-?: T[K] extends V ? never : K;
}[keyof T]>;
declare type FilterArray<A extends any[], V> = A extends [] ? [] : A extends [infer H, ...infer T] ? H extends V ? [H, ...FilterArray<T, V>] : FilterArray<T, V> : [];
declare type FilterNotArray<A extends any[], V> = A extends [] ? [] : A extends [infer H, ...infer T] ? H extends V ? FilterNotArray<T, V> : [H, ...FilterNotArray<T, V>] : [];
export declare type Merge<S, T> = Is<S, T> extends true ? S : Is<T, void> extends true ? S : Or<Is<S, never>, Is<T, never>> extends true ? never : Is<S, unknown> extends true ? unknown : Is<T, unknown> extends true ? S : Is<S, any> extends true ? any : Is<T, any> extends true ? S : S extends any[] ? (T extends any[] ? MergeArrays<S, T> : (S extends T ? S : never)) : S extends Fn ? (T extends Fn ? MergeFunctions<S, T> : (S extends T ? S : never)) : S extends Obj ? (T extends Obj ? MergeObjects<S, T> : (S extends T ? S : never)) : S extends T ? S : never;
declare type MergeArrays<S extends any[], T extends any[]> = S extends [infer HeadS, ...infer TailS] ? T extends [infer HeadT, ...infer TailT] ? [Merge<HeadS, HeadT>, ...MergeArrays<TailS, TailT>] : [never] : [];
declare type MergeFunctions<S extends Fn, T extends Fn> = S extends (...args: infer ArgsS) => infer ResS ? T extends (...args: infer ArgsT) => infer ResT ? Merge<ArgsS, ArgsT> extends infer A ? A extends any[] ? (...args: A) => Merge<ResS, ResT> : never : never : never : never;
export declare type MergeObjects<S, T> = Flatten<{
    [K in keyof S | keyof T]: K extends keyof S ? (K extends keyof T ? Merge<S[K], T[K]> : S[K]) : (K extends keyof T ? T[K] : never);
}>;
export declare type Values<T> = T[keyof T];
export {};
//# sourceMappingURL=objects.d.ts.map